import 'package:path/path.dart' as path;
import 'dart:collection';
import 'dart:core';
import 'dart:io';
import 'dart:async';

import 'dart:math' as math;

import 'package:io/ansi.dart';
import 'package:source_span/source_span.dart';

import '../helpers/logging.dart';
import '../modifiers/patch.dart';

class FileManager {
  Future<List<FileSystemEntity>> dirContentsAsync(String dirPath, {bool recursive = false, String extensionFilter}) {
    var dir = Directory(dirPath);
    var files = <FileSystemEntity>[];
    var completer = Completer();
    var lister = dir.list(recursive: recursive);
    lister.listen((file) {
      if (extensionFilter != null) {
        if (extensionFilter.isNotEmpty) {
          if (!isFileSelected(file.uri.path, extensionFilter: extensionFilter)) {
            return files.add(file);
          }
        }
      } else {
        return files.add(file);
      }
    },
        // should also register onError
        onDone: () => completer.complete(files));
    return completer.future;
  }

  List<FileSystemEntity> dirContentsSync(String dirPath, {bool recursive = false, String extensionFilter}) {
    var dir = Directory(dirPath);
    var list = dir.listSync(recursive: recursive);
    final List<FileSystemEntity> result = List<FileSystemEntity>();
    result.addAll(list);

    if (extensionFilter != null && list != null) {
      if (extensionFilter.isNotEmpty) {
        for (var item in list) {
          if (!isFileSelected(item.uri.path, extensionFilter: extensionFilter)) {
            result.remove(item);
          }
        }
      }
    }
    return result;
  }

  bool isFileSelected(String filePath, {String extensionFilter = '.dart'}) {
    return path.extension(filePath) == extensionFilter;
  }

  /// Applies all of the [patches] (insertions/deletions/replacements) to the
  /// contents of [sourceFile] and writes the result to disk.
  ///
  /// Throws an [ArgumentError] if [sourceFile] has a null value for
  /// [SourceFile.url], as it is required to open the file and write the new
  /// contents.
  void applyPatchesAndSave(SourceFile sourceFile, Iterable<Patch> patches) {
    if (patches.isEmpty) {
      return;
    }
    if (sourceFile.url == null) {
      throw ArgumentError('sourceFile.url cannot be null');
    }
    final updatedContents = applyPatches(sourceFile, patches);
    var filePath = sourceFile.url.toFilePath();
    //print('applyPatchesAndSave updatedContents ${updatedContents}');
    File(filePath).writeAsStringSync(updatedContents);
  }

  /// Returns the result of applying all of the [patches]
  /// (insertions/deletions/replacements) to the contents of [sourceFile].
  ///
  /// Throws an [Exception] if any two of the given [patches] overlap.
  String applyPatches(SourceFile sourceFile, Iterable<Patch> patches) {
    final buffer = StringBuffer();
    final sortedPatches = patches.toList()..sort();

    var lastEdgeOffset = 0;
    for (final patch in sortedPatches) {
      if (patch.startOffset < lastEdgeOffset) {
        throw Exception('Codemod terminated due to overlapping patch.\n'
            'Overlapping patch: ${patch.toString()}\n'
            'Updated text: ${patch.updatedText}\n');
      }

      // Write unmodified text from end of last patch to beginning of this patch
      buffer.write(sourceFile.getText(lastEdgeOffset, patch.startOffset));
      // Write the patched text (and do nothing with the original text, which is
      // effectively the same as replacing it)
      buffer.write(patch.updatedText);

      lastEdgeOffset = patch.endOffset;
    }

    final lastUnmodifiedText = sourceFile.getText(lastEdgeOffset);
    if (lastUnmodifiedText != null) {
      buffer.write(lastUnmodifiedText);
    }

    return buffer.toString();
  }

  /*
  Future<List<String>> dirContents(Directory dir) {
    var filenames = <String>[];
    var completer = new Completer();
    var lister = dir.list();
    lister.onFile = (filename) => filenames.add(filename);
    // should also register onError
    lister.onDone = (_) => completer.complete(filenames);
    return completer.future;
  }
 */
}
