import 'package:io/ansi.dart';
import 'package:logging/logging.dart';
import 'package:stack_trace/stack_trace.dart';

/// Logger to be used for all error and debug logs in this package.
final Logger logger = Logger('codemod');

/// Returns a listener function for the [Logger.onRecord] stream that writes all
/// [LogRecord]s to the given [sink].
///
/// If [ansiOutputEnabled] is true, logs will be highlighted based on the
/// [Level] using ansi color codes.
///
/// If [verbose] is true, additional information will be included in the log
/// messages including stack traces, logger name, and extra newlines.
Function(LogRecord) logListener(
  StringSink sink, {
  bool ansiOutputEnabled,
  bool verbose,
}) =>
    (record) => overrideAnsiOutput(ansiOutputEnabled == true, () {
          _logListener(record, sink, verbose: verbose ?? false);
        });

void _logListener(LogRecord record, StringSink sink, {bool verbose}) {
  AnsiCode color;
  if (record.level < Level.WARNING) {
    color = cyan;
  } else if (record.level < Level.SEVERE) {
    color = yellow;
  } else {
    color = red;
  }
  final level = color.wrap('[${record.level}]');
  var headerMessage = record.message;
  var blankLineCount = 0;
  if (headerMessage.startsWith('\n')) {
    blankLineCount = headerMessage.split('\n').takeWhile((line) => line.isEmpty).length;
    headerMessage = headerMessage.substring(blankLineCount);
  }
  final header = '$level ${_loggerName(record, verbose)}$headerMessage';
  final lines = blankLineCount > 0 ? (List<Object>.generate(blankLineCount, (_) => '')..add(header)) : <Object>[header];

  if (record.error != null) {
    lines.add(record.error);
  }

  if (record.stackTrace != null && verbose) {
    lines.add(Trace.from(record.stackTrace).terse);
  }

  sink.writeln(lines.join('\n'));
}

/// Filter out the Logger names known to come from `codemod` and splits the
/// header for levels >= WARNING.
String _loggerName(LogRecord record, bool verbose) {
  final knownNames = const [
    'codemod',
  ];
  return verbose || !knownNames.contains(record.loggerName) ? '${record.loggerName}: ' : '';
}
